"""File for defining BotBlox CLI."""

import argparse
import logging
import sys
import time
from typing import List

import serial

from .data_manager import (
    portmirror_create_configuration,
    vlan_create_configuration,
)

logging.basicConfig(level=logging.DEBUG)


def write_data_to_serial(
    data: List[List],
    device_name: str,
) -> bool:
    """Write data commands to serial port.

    Write data to serial port which the UART converter is connected to. The STM32 MCU
    on the SwitchBlox will then be interrupted and carry out the commands

    :param List[List] data: Commands created to write to STM32 MCU
    :param Optional[str] device_name: serial port device name (default: '/dev/ttyUSB0')
    :return: Flag to indicate whether the write data to serial was successful or not
    :rtype: bool
    """
    ser = serial.Serial(
        port=device_name,
        baudrate=115200,
        bytesize=serial.EIGHTBITS,
        parity=serial.PARITY_NONE,
        timeout=20,
        write_timeout=2,
    )

    for command in data:
        x = bytes(command)
        ser.write(x)
        time.sleep(0.1)

    condition = ser.read(size=1)
    ser.close()

    try:
        condition = list(condition)[0]
    except IndexError:
        logging.error('Failed to read condition message from board')
        return False
    else:
        if condition == 1:
            logging.info('Success setting configuration in EEPROM')
            return True
        elif condition == 2:
            logging.error('Failed saving configuration in EEPROM')
            return False


def cli() -> None:
    """Define all cli parser and subparsers here."""
    parser = argparse.ArgumentParser(
        description='CLI for configuring SwitchBlox managed settings',
        epilog='Please open any issue on https://github.com/botblox/botblox-manager-software/ if there is a problem',
    )
    parser.add_argument(
        '-D',
        '--device',
        type=str,
        help='Select the USB-to-UART converter device',
        nargs=1,
        required=True,
    )

    subparsers = parser.add_subparsers(
        title='Individual group commands for each configuration',
        description='Please choose a certain command',
        required=True,
    )

    # VLAN CLI
    vlan_parser = subparsers.add_parser(
        'vlan',
        help='Configure the ports to be in VLAN groups',
    )
    vlan_parser_group = vlan_parser.add_mutually_exclusive_group()
    vlan_parser_group.add_argument(
        '-g',
        '--group',
        nargs='+',
        action='append',
        type=int,
        choices=[1, 2, 3, 4, 5],
        required=False,
        help='''Define the VLAN member groups using port number,
        i.e. --group 1 2 --group 3 4 puts makes Group A have
        ports 1 and 2, and Group B have ports 3 and 4'''
    )
    vlan_parser_group.add_argument(
        '-r',
        '--reset',
        action='store_true',
        default=argparse.SUPPRESS,
        help='''Reset the VLAN configuration to be as system default'''
    )
    vlan_parser_group.set_defaults(execute=vlan_create_configuration)

    # Port mirroring CLI
    portmirror_parser = subparsers.add_parser(
        'mirror',
        help='Configure the ports to mirror traffic',
    )
    portmirror_parser_mutex_grouping = portmirror_parser.add_mutually_exclusive_group()
    portmirror_parser_config_group = portmirror_parser_mutex_grouping.add_argument_group()

    portmirror_parser_config_group.add_argument(
        '-m',
        '--mode',
        nargs='?',
        default='RX',
        type=str,
        choices=['RX', 'TX', 'RXorTX', 'RXandTX'],
        required=False,
        help='''Select the mirror mode during operation''',
    )
    portmirror_parser_config_group.add_argument(
        '-M',
        '--mirror-port',
        nargs=1,
        type=int,
        choices=[1, 2, 3, 4, 5],
        required=False,
        default=5,
        help='''Select the mirror port (default: port 5)''',
    )
    portmirror_parser_config_group.add_argument(
        '-rx',
        '--rx-port',
        nargs='+',
        type=int,
        choices=[1, 2, 3, 4, 5],
        required=any(mode in sys.argv for mode in ['RX', 'RXorTX', 'RXandTX']),
        default=argparse.SUPPRESS,
        help='''Select the source (receive) port to be mirrored''',
    )
    portmirror_parser_config_group.add_argument(
        '-tx',
        '--tx-port',
        nargs='+',
        type=int,
        choices=[1, 2, 3, 4, 5],
        required=any(mode in sys.argv for mode in ['TX', 'RXorTX', 'RXandTX']),
        default=argparse.SUPPRESS,
        help='''Select the destination (transmit) port to be mirrored''',
    )
    portmirror_parser_mutex_grouping.add_argument(
        '-r',
        '--reset',
        action='store_true',
        default=argparse.SUPPRESS,
        help='Reset the Port mirroring configuration to default, this will turn port mirroring off'
    )
    portmirror_parser_mutex_grouping.set_defaults(execute=portmirror_create_configuration)

    if len(sys.argv) <= 1:
        sys.argv.append('--help')

    args = parser.parse_args()
    data = args.execute(args)

    # add stop command
    data.append([100, 0, 0, 0])

    device_name = args.device_name
    is_success = write_data_to_serial(data=data, device_name=device_name)

    if is_success:
        logging.info('Successful configuration')
    else:
        logging.error('Failed to configure - check logs')



"""Define helper functions to convert the user vlan configuration to commands for the firmware."""

from typing import (
    List,
    Tuple,
    Union,
)

vlan_miim_register_map = {
    1: {
        'name': 'PBV_MEMBER_P0',
        'phy': 23,
        'reg': 16,
        'size': 8,
        'offset': 0,
        'sys_default': 0xFF,
        'data': 0,
    },
    2: {
        'name': 'PBV_MEMBER_P1',
        'phy': 23,
        'reg': 16,
        'size': 8,
        'offset': 8,
        'sys_default': 0xFF,
        'data': 0,
    },
    3: {
        'name': 'PBV_MEMBER_P2',
        'phy': 23,
        'reg': 17,
        'size': 8,
        'offset': 0,
        'sys_default': 0xFF,
        'data': 0,
    },
    4: {
        'name': 'PBV_MEMBER_P3',
        'phy': 23,
        'reg': 18,
        'size': 8,
        'offset': 0,
        'sys_default': 0xFF,
        'data': 0,
    },
    5: {
        'name': 'PBV_MEMBER_P4',
        'phy': 23,
        'reg': 18,
        'size': 8,
        'offset': 8,
        'sys_default': 0xFF,
        'data': 0,
    },
}

vlan_binary_offset_map = {
        1: 2,
        2: 3,
        3: 4,
        4: 6,
        5: 7,
}


def get_unique_ports(
    numbers: List,
) -> List:
    """Get unique member ports.

    Returns only the unique ports that a single port shares VLAN membership with

    :param List numbers: List of ports a single port shares VLAN membership, duplicates allowed.
    :return: List of ports a single port shares VLAN membership, duplicates forbidden.
    :rtype: List

    :example:

    >>> numbers=[1, 1, 2, 3, 4]
    >>> get_unique_ports(numbers)
    [1, 2, 3, 4]
    """
    unique = []
    for number in numbers:
        if number not in unique:
            unique.append(number)
    return unique


def create_vlan_parser_data_per_port(
    vlan_members: List,
    port: int,
) -> None:
    """Create section of parser data command per port.

    Loops over members ports for each physical port and create the data in the shape so that it can be used
    by the parser when further checks on duplicate commands are made.

    :param List vlan_members: List of the member ports belonging to the physical port
    :param int port: individual physical port
    :rtype: None
    """
    for other_port in vlan_members:
        vlan_miim_register_map[port]['data'] = vlan_miim_register_map[port]['data'] | (
            1 << vlan_binary_offset_map[other_port])


def is_command_already_present(
    data: List[Union[List, None]],
    port: int,
) -> Tuple[bool, Union[int, None]]:
    """Calculate if a command writing to same PHY and REG exists.

    Loops over current commands and determines if a command with the same PHY and REG is present.
    If so, return a tuple with (True, <index_of_command_in_data>).
    Else, return a tuple with (False, None)

    :param List[Union[List, None]] data: Current commands created, nullable input.
    :param int port: individual physical port
    :return: object with boolean flag and index representing position in list where command already exists
        for same PHY and REG
    :rtype: Tuple[bool, Union[int, None]]

    :example:

    >>> data=[[23, 16, 0, 0]]
    >>> port=2
    >>> is_command_already_present(data, port)
    (True, 0)
    """
    if data:
        for index_in_data, command in enumerate(data):
            if vlan_miim_register_map[port]['phy'] == command[0] and vlan_miim_register_map[port]['reg'] == command[1]:
                return (True, index_in_data)

    return (False, None)


def update_command(
    command_to_update: List,
    port: int,
) -> List:
    """Update a command when it is found that two parts of data share the same PHY and REG.

    Use the `vlan_miim_register_map` to update the command with the register offset of the
    new data that shared the same PHY and REG.

    :param List command_to_update: Current command that needs to be updated with new data.
    :param int port: individual physical port
    :return: Updated command with data regarding the current port
    :rtype: List
    """
    port_data = 0
    if vlan_miim_register_map[port]['data'] != 0:
        port_data = vlan_miim_register_map[port]['data']
    else:
        port_data = vlan_miim_register_map[port]['sys_default']
    port_offset = vlan_miim_register_map[port]['offset']

    command_to_update[2] = command_to_update[2] | ((port_data << port_offset) & 0xFF)
    command_to_update[3] = command_to_update[3] | (((port_data << port_offset) & 0xFF00) >> port_offset)

    return command_to_update


def create_command(
    port: int,
) -> List:
    """Create a new command with PHY and REG plus two bytes of command data.

    Use the `vlan_miim_register_map` to create the command

    :param int port: individual physical port
    :return: Newly created command
    :rtype: List
    """
    port_data = 0
    if vlan_miim_register_map[port]['data'] != 0:
        port_data = vlan_miim_register_map[port]['data']
    else:
        port_data = vlan_miim_register_map[port]['sys_default']
    port_offset = vlan_miim_register_map[port]['offset']

    command = [
        vlan_miim_register_map[port]['phy'],
        vlan_miim_register_map[port]['reg'],
        (port_data << port_offset) & 0xFF,
        ((port_data << port_offset) & 0xFF00) >> port_offset,
    ]
    return command


def update_data(
    data: List[List],
    port: int,
) -> List[List]:
    """Update data commands for setting the VLAN configuration.

    If this function is executed before the `vlan_miim_register_map` is filled out,
    this will return system defaults

    :param List[List] data: current list of commands
    :param int port: physical port
    :return: Command for physical port
    :rtype: List[List]
    """
    is_present, index_in_data = is_command_already_present(data, port)

    if is_present and index_in_data is not None:
        command_to_update = data[index_in_data]
        command_updated = update_command(command_to_update=command_to_update, port=port)
        data[index_in_data] = command_updated
    else:
        command = create_command(port)
        data.append(command)

    return data


def vlan_create_configuration(
    args: Tuple,
) -> List[List]:
    """Create parser data from user input.

    :param Tuple args: command line arguments from user
    :return: list of commmands to be sent to the firmware
    :rtype: List[List]

    :example:

    >>> args=(execute=<function>, group=[[1, 2], [3, 4]])
    >>> vlan_create_configuration(groupings)
    [[23, 16, 12, 12], [23, 17, 80, 0], [23, 18, 80, 0]]

    >>> args=(execute=<function>, reset=True)
    >>> vlan_create_configuraton(groupings)
    [[23, 16, 255, 255], [23, 17, 255, 0], [23, 18, 255, 255]]
    """
    data = []
    ports = list(vlan_binary_offset_map.keys())
    try:
        args.reset
    except AttributeError:
        groupings = args.group
    else:
        for port in ports:
            data = update_data(data=data, port=port)
        return data

    for port in ports:
        vlan_members = []
        for grouping in groupings:
            if port in grouping:
                vlan_members += grouping

        vlan_members = get_unique_ports(vlan_members)

        create_vlan_parser_data_per_port(vlan_members=vlan_members, port=port)

        data = update_data(data=data, port=port)

    return data



"""Package configuration."""

import os
import pathlib

from manager import __version__
from setuptools import (
    find_packages,
    setup,
)

HERE = pathlib.Path(__file__).parent

# README = (HERE / 'README.md').read_text()
readme_path = os.path.join(HERE, 'README.md')
README = open(readme_path, 'rt').read()

setup(
    name='botblox',
    version=__version__,
    description='BotBlox Manager CLI tool',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/botblox/botblox-manager-software',
    author='Aaron Elijah',
    author_email='aaronzakelijah@googlemail.com',
    license='MIT',
    python_requires='>=3.6',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Operating System :: OS Independent',
    ],
    packages=find_packages(include=['manager', 'manager.*']),
    include_package_data=True,
    install_requires=[
        'pyserial>=3.5',
    ],
    extras_require={
        'dev': [
            'flake8==3.8.4',
            'flake8-import-order==0.18.1',
            'flake8-builtins==1.5.3',
            'flake8-docstrings==1.5.0',
            'flake8-rst-docstrings==0.0.14',
            'flake8-annotations==2.4.1',
            'flake8-print==4.0.0',
            'pep8-naming==0.11.1',
        ],
    },
    entry_points={
        'console_scripts': [
            'botblox=manager.__main__:main',
        ],
    },
)



"""Define helper functions to convert the user port mirror configuration to commands for the firmware."""

from typing import (
    List,
    Optional,
    Tuple,
    Union,
)

portmirror_miim_register_map = {
    'enable': {
        'name': 'PORT_MIRROR_EN',
        'phy': 20,
        'reg': 3,
        'size': 1,
        'offset': 15,
        'sys_default': 0,
        'data': -1,
    },
    'mode': {
        'name': 'PORT_MIRROR_MODE',
        'phy': 20,
        'reg': 3,
        'size': 2,
        'offset': 13,
        'sys_default': 0,
        'data': -1,
        'choice_mapping': {
            'RX': 0b00,
            'TX': 0b01,
            'RXandTX': 0b10,
            'RXorTX': 0b11,
        },
    },
    'rx_port': {
        'name': 'SEL_RX_PORT_MIRROR',
        'phy': 20,
        'reg': 3,
        'size': 8,
        'offset': 0,
        'sys_default': 0x01,
        'data': -1,
        'choice_mapping': {
            1: 0b00000100,
            2: 0b00001000,
            3: 0b00010000,
            4: 0b01000000,
            5: 0b10000000,
        },
    },
    'mirror_port': {
        'name': 'SEL_MIRROR_PORT',
        'phy': 20,
        'reg': 4,
        'size': 3,
        'offset': 13,
        'sys_default': 0b111,
        'data': -1,
        'choice_mapping': {
            1: 0b010,
            2: 0b011,
            3: 0b100,
            4: 0b110,
            5: 0b111,
        },
    },
    'tx_port': {
        'name': 'SEL_TX_PORT_MIRROR',
        'phy': 20,
        'reg': 4,
        'size': 8,
        'offset': 0,
        'sys_default': 0x01,
        'data': -1,
        'choice_mapping': {
            1: 0b00000100,
            2: 0b00001000,
            3: 0b00010000,
            4: 0b01000000,
            5: 0b10000000,
        },
    },
}


def create_portmirror_parser_data_per_option(
    option: str,
    settings: Optional[Union[List[int], str, int]],
) -> None:
    """Create section of parser data command for port mirror option.

    For a given option for port mirror, link the user input on the command line with
    the given choices corresponding to that input from the `portmirror_miim_register_map`.
    Some choices for options allow multiple choices to be selected and so each choice must
    be bitwise 'OR'-ed with each other to create the parser data for that option.
    For 'enable', given that no user explicity specifies that as an arg, it is automatically
    assigned the apprioriate value for activated port mirroring.

    :param str option: individual port mirror option
    :param Optional[Union[List[int], str, int]] settings: Can be a string relating to the port mirror mode (i.e. 'RX'),
        int relating to mirror_port (i.e. 1) or
        List[int] of multiple choices for other options (i.e. [1, 2] for `tx_port`)
    :rtype: None
    """
    if option == 'enable':
        portmirror_miim_register_map[option]['data'] = 1
        return

    choice_mapping = portmirror_miim_register_map[option]['choice_mapping']
    if type(settings) == list:
        individual_setting_data = [choice_mapping[setting] for setting in settings]
    else:
        individual_setting_data = [choice_mapping[settings]]

    data = 0
    if len(individual_setting_data) > 1:
        for setting_data in individual_setting_data:
            data = data | setting_data
    else:
        data = individual_setting_data[0]

    portmirror_miim_register_map[option]['data'] = data


def is_command_already_present(
    data: List[List],
    option: str,
) -> Tuple[bool, Union[int, None]]:
    """Calculate if a command writing to same PHY and REG already exists in data.

    Loops over current commands in data and determines if a command with the same PHY and REG is present.
    If so, return a tuple with (True, <index_of_command_in_data>).
    Else, return a tuple with (False, None)

    :param List[Union[List, None]] data: Current commands created, nullable input.
    :param str option: individual port mirror option
    :return: object with boolean flag and index representing position in list where command already exists
        for same PHY and REG
    :rtype: Tuple[bool, Union[int, None]]

    :example:

    >>> data=[[20, 3, 0, 128]]
    >>> option='mode'
    >>> is_command_already_present(data, option)
    (True, 0)
    """
    if data:
        for index_in_data, command in enumerate(data):
            is_command_with_same_phy = portmirror_miim_register_map[option]['phy'] == command[0]
            is_command_with_same_reg = portmirror_miim_register_map[option]['reg'] == command[1]
            if is_command_with_same_phy and is_command_with_same_reg:
                return (True, index_in_data)

    return (False, None)


def update_command(
    command_to_update: List,
    option: str,
) -> List:
    """Update a command when it is found that two parts of data share the same PHY and REG.

    Use the `portmirror_miim_register_map` to update the command with the register offset of the
    new data that shared the same PHY and REG. If two user inputs refer to data
    that share the same PHY and REG, then it needs to be part of the same command.

    :param List command_to_update: Current command that needs to be updated with new data.
    :param str option: individual port mirror option
    :return: Updated command with data regarding the current port
    :rtype: List
    """
    if portmirror_miim_register_map[option]['data'] >= 0:
        option_data = portmirror_miim_register_map[option]['data']
    else:
        option_data = portmirror_miim_register_map[option]['sys_default']
    option_offset = portmirror_miim_register_map[option]['offset']

    command_to_update[2] = command_to_update[2] | ((option_data << option_offset) & 0xFF)
    command_to_update[3] = command_to_update[3] | (((option_data << option_offset) & 0xFF00) >> 8)

    return command_to_update


def create_command(
    option: str,
    reset: Optional[bool] = False,
) -> List:
    """Create a new command with PHY and REG plus two bytes of command data.

    Use the `portmirror_miim_register_map` to create the command.
    If reset is defined in user command line arguments, then we reset the
    port mirror to be system defaults. Elsewise use data in the
    `portmirror_miim_register_map`.

    :param str option: individual port mirror option
    :return: New created command
    :rtype: List
    """
    if reset:
        option_data = portmirror_miim_register_map[option]['sys_default']
    else:
        option_data = portmirror_miim_register_map[option]['data']
    option_offset = portmirror_miim_register_map[option]['offset']

    command = [
        portmirror_miim_register_map[option]['phy'],
        portmirror_miim_register_map[option]['reg'],
        (option_data << option_offset) & 0xFF,
        ((option_data << option_offset) & 0xFF00) >> 8,
    ]
    return command


def update_data(
    data: List[List],
    option: str,
    reset: Optional[bool] = False,
) -> List[List]:
    """Update data commands for setting the port mirror configuration.

    If this function is executed before the `portmirror_miim_register_map` is filled out,
    this will return system defaults for port mirroing

    :param List[List] data: current list of commands
    :param int port: physical port
    :return: Command for physical port
    :rtype: List[List]
    """
    is_present, index_in_data = is_command_already_present(data, option)

    if is_present and index_in_data is not None:
        command_to_update = data[index_in_data]
        command_updated = update_command(command_to_update=command_to_update, option=option)
        data[index_in_data] = command_updated
    else:
        command = create_command(option=option, reset=reset)
        data.append(command)

    return data


def portmirror_create_configuration(
    args: Tuple,
) -> List[List]:
    """Create parser data from user input.

    :param Tuple args: command line arguments from user
    :return: list of commmands to be sent to the firmware
    :rtype: List[List]

    :example:

    >>> args=Namespace(execute=<function portmirror_create_configuration at {pointer}>,
                       mirror_port=[4], mode='RXandTX', rx_port=[1, 2], tx_port=[3, 5])
    >>> portmirror_create_configuration(args)
    [[20, 3, 12, 192], [20, 4, 144, 192]]

    >>> args=Namespace(execute=<function portmirror_create_configuration at {pointer}>,
                       mirror_port=5, mode='RX', reset=True)
    >>> portmirror_create_configuration(args)
    [[20, 3, 1, 0], [20, 4, 1, 224]]
    """
    data = []
    mirror_options = list(portmirror_miim_register_map.keys())
    try:
        args.reset
    except AttributeError:
        args_dict = vars(args)
    else:
        for mirror_option in mirror_options:
            data = update_data(data=data, option=mirror_option, reset=True)
        return data

    for mirror_option in mirror_options:
        if mirror_option == 'enable':
            create_portmirror_parser_data_per_option(option=mirror_option, settings=1)
        else:
            if args_dict.get(mirror_option, None) is not None:
                create_portmirror_parser_data_per_option(option=mirror_option, settings=args_dict[mirror_option])

        data = update_data(data=data, option=mirror_option)

    return data



"""Package for containing data manager functions to convert user input to commands."""

from .portmirror_config import portmirror_create_configuration
from .vlan_config import vlan_create_configuration

__all__ = [
    'portmirror_create_configuration',
    'vlan_create_configuration',
]
